/*
 * Export Letux 400 display operations as a server.
 *
 * Copyright (C) 2018, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/gpio-jz4730.h>
#include <l4/devices/memory.h>

#include <l4/re/env.h>

#include <ipc/server.h>
#include "activation_client.h"
#include "activation_server.h"

/* Virtual address for the GPIO register block. */

static l4_addr_t gpio_virt_base = 0, gpio_virt_base_end = 0;

/* GPIO abstraction. */

static Gpio_jz4730_chip *gpio_port_b = 0, *gpio_port_c = 0;



/* GPIO pin definitions. */

enum Jz4730_lcd_gpio_mask
{
  Jz4730_lcd_gpio_func1 = 0xf8ffff00,
  Jz4730_lcd_gpio_func2 = 0x07000000,
};

enum Jz4730_lcd_gpio
{
  Jz4730_lcd_gpio_led_enable     = 28,
  Jz4730_lcd_gpio_display_enable = 29,
};



static int setup_memory(void)
{
  if (get_memory("jz4730-gpio", &gpio_virt_base, &gpio_virt_base_end))
    return 1;

  return 0;
}



/* Display device only implementing the activation interface. */

class server_DisplayObject : public Activation
{
  Pin_slice lcd_mask1 = {.offset=0, .mask=Jz4730_lcd_gpio_func1};
  Pin_slice lcd_mask2 = {.offset=0, .mask=Jz4730_lcd_gpio_func2};

  Activation *_backlight;

public:
  explicit server_DisplayObject(Activation *backlight)
  : _backlight(backlight)
  {
    /* Set functions for the LCD pins. */

    gpio_port_b->multi_config_pad(lcd_mask1, Hw::Gpio_chip::Function_alt, 1);
    gpio_port_b->multi_config_pad(lcd_mask2, Hw::Gpio_chip::Function_alt, 2);

    /* Enable LED, disable display. */

    gpio_port_c->setup(Jz4730_lcd_gpio_led_enable, Hw::Gpio_chip::Output, 1);
    gpio_port_c->setup(Jz4730_lcd_gpio_display_enable, Hw::Gpio_chip::Output, 0);
  }

  /* Switch the display off. */

  long disable(void)
  {
    gpio_port_c->setup(Jz4730_lcd_gpio_display_enable, Hw::Gpio_chip::Output, 0);
    _backlight->disable();
    return L4_EOK;
  }

  /* Switch the display on. */

  long enable(void)
  {
    gpio_port_c->setup(Jz4730_lcd_gpio_display_enable, Hw::Gpio_chip::Output, 1);
    _backlight->enable();
    return L4_EOK;
  }
};



int main(void)
{
  if (setup_memory()) return 1;

  /* Initialise the GPIO abstractions. */

  Gpio_jz4730_chip gpb(gpio_virt_base + (1 * 0x30), gpio_virt_base + (2 * 0x30), 32);
  Gpio_jz4730_chip gpc(gpio_virt_base + (2 * 0x30), gpio_virt_base + (3 * 0x30), 32);

  gpio_port_b = &gpb;
  gpio_port_c = &gpc;

  /* Obtain a reference to the backlight device. */

  l4_cap_idx_t backlight = l4re_env_get_cap("backlight");
  if (!l4_is_valid_cap(backlight)) return 1;

  /* Use the activation interface with the backlight. */

  client_Activation backlight_obj(backlight);

  /* Initialise and register a new server object. */

  server_DisplayObject obj(&backlight_obj);
  l4_cap_idx_t server;

  if (ipc_server_bind("display", (l4_umword_t) &obj, &server)) return 1;

  /* Enter the IPC server loop. */

  ipc_server_loop(Activation_expected_items, &obj,
                  (ipc_server_handler_type) handle_Activation);
  return 0;
}
